# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the crypto registry."""

import logging
import string
from unittest import mock

import pytest
from aea_ledger_cosmos import CosmosCrypto
from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_fetchai import FetchAICrypto

import aea.crypto
from aea.crypto.registries.base import EntryPoint
from aea.exceptions import AEAException

from tests.data.custom_crypto import CustomCrypto


logger = logging.getLogger(__name__)

forbidden_special_characters = "".join(
    filter(lambda c: c not in "_:/.", string.punctuation)
)


def test_make_fetchai():
    """Test the 'make' method for 'fetchai' crypto."""
    aea_ledger_fetchai = aea.crypto.registries.make_crypto(FetchAICrypto.identifier)

    # calling 'make' again will give a different object.
    aea_ledger_fetchai_1 = aea.crypto.registries.make_crypto(FetchAICrypto.identifier)
    assert type(aea_ledger_fetchai) == type(aea_ledger_fetchai_1)
    assert aea_ledger_fetchai.address != aea_ledger_fetchai_1


def test_make_ethereum():
    """Test the 'make' method for 'ethereum' crypto."""
    aea_ledger_ethereum = aea.crypto.registries.make_crypto(EthereumCrypto.identifier)

    # calling 'make' again will give a different object.
    aea_ledger_ethereum_1 = aea.crypto.registries.make_crypto(EthereumCrypto.identifier)
    assert type(aea_ledger_ethereum) == type(aea_ledger_ethereum_1)
    assert aea_ledger_ethereum.address != aea_ledger_ethereum_1.address


def test_make_cosmos():
    """Test the 'make' method for 'cosmos' crypto."""
    aea_ledger_cosmos = aea.crypto.registries.make_crypto(CosmosCrypto.identifier)

    # calling 'make' again will give a different object.
    aea_ledger_cosmos_1 = aea.crypto.registries.make_crypto(CosmosCrypto.identifier)
    assert type(aea_ledger_cosmos) == type(aea_ledger_cosmos_1)
    assert aea_ledger_cosmos.address != aea_ledger_cosmos_1.address


def test_register_custom_crypto():
    """Test the 'register' method with a custom crypto object."""

    aea.crypto.registries.register_crypto(
        "my_custom_crypto", entry_point="tests.data.custom_crypto:CustomCrypto"
    )

    assert (
        aea.crypto.registries.crypto_registry.specs.get("my_custom_crypto") is not None
    )
    actual_spec = aea.crypto.registries.crypto_registry.specs["my_custom_crypto"]

    expected_id = "my_custom_crypto"
    expected_entry_point = EntryPoint("tests.data.custom_crypto:CustomCrypto")
    assert actual_spec.id == expected_id
    assert actual_spec.entry_point == expected_entry_point
    assert actual_spec.entry_point.import_path == expected_entry_point.import_path
    assert actual_spec.entry_point.class_name == expected_entry_point.class_name

    my_crypto = aea.crypto.registries.make_crypto("my_custom_crypto")
    assert type(my_crypto) == CustomCrypto

    # calling 'make' again will give a different object.
    my_crypto_1 = aea.crypto.registries.make_crypto("my_custom_crypto")
    assert type(my_crypto) == type(my_crypto_1)
    assert my_crypto != my_crypto_1

    aea.crypto.registries.crypto_registry.specs.pop("my_custom_crypto")


def test_cannot_register_crypto_twice():
    """Test we cannot register a crypto twice."""
    aea.crypto.registries.register_crypto(
        "my_custom_crypto", entry_point="tests.data.custom_crypto:CustomCrypto"
    )

    with pytest.raises(AEAException, match="Cannot re-register id: 'my_custom_crypto'"):
        aea.crypto.registries.register_crypto(
            "my_custom_crypto", entry_point="tests.data.custom_crypto:CustomCrypto"
        )

    aea.crypto.registries.crypto_registry.specs.pop("my_custom_crypto")


@mock.patch("importlib.import_module", side_effect=ImportError)
def test_import_error(*mocks):
    """Test import errors."""
    aea.crypto.registries.register_crypto(
        "some_crypto", entry_point="path.to.module:SomeCrypto"
    )
    with pytest.raises(
        AEAException,
        match="A module (.*) was specified for the item but was not found",
    ):
        aea.crypto.registries.make_crypto("some_crypto", module="some.module")
    aea.crypto.registries.crypto_registry.specs.pop("some_crypto")


class TestRegisterWithMalformedId:
    """Test the error message when we try to register a crypto whose identifier is malformed."""

    MESSAGE_REGEX = "Malformed .*: '.*'. It must be of the form '.*'."

    def test_wrong_spaces(self):
        """Spaces not allowed in a Crypto ID."""
        # beginning space
        with pytest.raises(AEAException, match=self.MESSAGE_REGEX):
            aea.crypto.registries.register_crypto(
                " malformed_id", "path.to.module:CryptoClass"
            )

        # trailing space
        with pytest.raises(AEAException, match=self.MESSAGE_REGEX):
            aea.crypto.registries.register_crypto(
                "malformed_id ", "path.to.module:CryptoClass"
            )

        # in between
        with pytest.raises(AEAException, match=self.MESSAGE_REGEX):
            aea.crypto.registries.register_crypto(
                "malformed id", "path.to.module:CryptoClass"
            )

    @pytest.mark.parametrize("special_character", forbidden_special_characters)
    def test_special_characters(self, special_character):
        """Special characters are not allowed (only underscore)."""
        with pytest.raises(AEAException, match=self.MESSAGE_REGEX):
            aea.crypto.registries.register_crypto(
                "malformed_id" + special_character, "path.to.module:CryptoClass"
            )


class TestRegisterWithMalformedEntryPoint:
    """Test the error message when we try to register a crypto with a wrong entry point."""

    MESSAGE_REGEX = "Malformed .*: '.*'. It must be of the form '.*'."

    def test_wrong_spaces(self):
        """Spaces not allowed in a Crypto ID."""
        # beginning space
        with pytest.raises(AEAException, match=self.MESSAGE_REGEX):
            aea.crypto.registries.register_crypto(
                "crypto_id", " path.to.module:CryptoClass"
            )

        # trailing space
        with pytest.raises(AEAException, match=self.MESSAGE_REGEX):
            aea.crypto.registries.register_crypto(
                "crypto_id", "path.to.module :CryptoClass"
            )

        # in between
        with pytest.raises(AEAException, match=self.MESSAGE_REGEX):
            aea.crypto.registries.register_crypto(
                "crypto_id", "path.to .module:CryptoClass"
            )

    @pytest.mark.parametrize("special_character", forbidden_special_characters)
    def test_special_characters(self, special_character):
        """Special characters are not allowed (only underscore)."""
        with pytest.raises(AEAException, match=self.MESSAGE_REGEX):
            aea.crypto.registries.register_crypto(
                "crypto_id", "path" + special_character + ".to.module:CryptoClass"
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the ethereum module."""

import pytest
from aea_ledger_ethereum import EthereumApi

from tests.conftest import MAX_FLAKY_RERUNS


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_contract_instance(erc1155_contract, ethereum_testnet_config):
    """Test the get contract instance method."""
    contract, contract_address = erc1155_contract
    ethereum_api = EthereumApi(**ethereum_testnet_config)
    interface = {"abi": [], "bytecode": b""}
    instance = ethereum_api.get_contract_instance(
        contract_interface=interface, contract_address=contract_address,
    )
    assert str(type(instance)) == "<class 'web3._utils.datatypes.Contract'>"
    instance = ethereum_api.get_contract_instance(contract_interface=interface,)
    assert (
        str(type(instance)) == "<class 'web3._utils.datatypes.PropertyCheckingFactory'>"
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the FET ERC20 contract definition."""

import logging

from aea_ledger_ethereum import EthereumApi

from aea.common import Address
from aea.configurations.base import PublicId
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi


_default_logger = logging.getLogger("aea.packages.fetchai.contracts.fet_erc20.contract")

PUBLIC_ID = PublicId.from_str("fetchai/fet_erc20:0.4.0")


class FetERC20(Contract):
    """The FetERC20 contract class which acts as a bridge between AEA framework and ERC20 ABI."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_approve_transaction(
        cls,
        ledger_api: LedgerApi,
        contract_address: Address,
        from_address: Address,
        spender: Address,
        amount: int,
        gas: int = 0,
    ) -> None:
        """
        Get transaction to approve oracle client contract transactions on behalf of sender.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param from_address: the address of the approver.
        :param spender: the address approved to spend on behalf of sender.
        :param amount: the amount approved to be spent.
        :param gas: the gas limit for the transaction.
        :return: None
        """
        if ledger_api.identifier == EthereumApi.identifier:
            nonce = ledger_api.api.eth.getTransactionCount(from_address)
            instance = cls.get_instance(ledger_api, contract_address)
            function = instance.functions.approve
            intermediate = function(spender, amount)
            tx = intermediate.buildTransaction(
                {
                    "gas": gas,
                    "gasPrice": ledger_api.api.toWei("50", "gwei"),
                    "nonce": nonce,
                }
            )
            tx = ledger_api.update_with_gas_estimate(tx)
            return tx
        raise NotImplementedError

    @classmethod
    def get_transfer_transaction(
        cls,
        ledger_api: LedgerApi,
        contract_address: Address,
        from_address: Address,
        receiver: Address,
        amount: int,
        gas: int = 0,
    ) -> None:
        """
        Get transaction to transfer tokens to an account

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param from_address: the address of the sender.
        :param receiver: the address to which to transfer tokens.
        :param amount: the amount of tokens to transfer.
        :param gas: the gas limit for the transaction.
        :return: None
        """
        if ledger_api.identifier == EthereumApi.identifier:
            nonce = ledger_api.api.eth.getTransactionCount(from_address)
            instance = cls.get_instance(ledger_api, contract_address)
            function = instance.functions.transfer
            intermediate = function(receiver, amount)
            tx = intermediate.buildTransaction(
                {
                    "gas": gas,
                    "gasPrice": ledger_api.api.toWei("50", "gwei"),
                    "nonce": nonce,
                }
            )
            tx = ledger_api.update_with_gas_estimate(tx)
            return tx
        raise NotImplementedError



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Utils used for operating Registry with CLI."""

import os
import tarfile
from json.decoder import JSONDecodeError
from typing import Any, Callable, Dict, Optional, Tuple, Union, cast

import click

from aea.cli.registry.settings import AUTH_TOKEN_KEY, REGISTRY_API_URL
from aea.cli.utils.config import get_or_create_cli_config
from aea.cli.utils.context import Context
from aea.cli.utils.loggers import logger
from aea.cli.utils.package_utils import find_item_locally
from aea.common import JSONLike
from aea.configurations.base import PublicId
from aea.helpers import http_requests as requests


FILE_DOWNLOAD_TIMEOUT = (
    180  # quite big number case possible slow channels and package can be quite big
)


def get_auth_token() -> str:
    """
    Get current auth token.

    :return: str auth token
    """
    config = get_or_create_cli_config()
    return config.get(AUTH_TOKEN_KEY, None)


def request_api(
    method: str,
    path: str,
    params: Optional[Dict] = None,
    data: Optional[Dict] = None,
    is_auth: bool = False,
    files: Optional[Dict] = None,
    handle_400: bool = True,
    return_code: bool = False,
) -> Union[JSONLike, Tuple[JSONLike, int]]:
    """
    Request Registry API.

    :param method: str request method ('GET, 'POST', 'PUT', etc.).
    :param path: str URL path.
    :param params: dict GET params.
    :param data: dict POST data.
    :param is_auth: bool is auth requied (default False).
    :param files: optional dict {file_field_name: open(filepath, "rb")} (default None).

    :return: dict response from Registry API or tuple (dict response, status code).
    """
    headers = {}
    if is_auth:
        token = get_auth_token()
        if token is None:
            raise click.ClickException(
                "Unable to read authentication config. "
                'Please sign in with "aea login" command.'
            )
        headers.update({"Authorization": "Token {}".format(token)})

    request_kwargs = dict(
        method=method,
        url="{}{}".format(REGISTRY_API_URL, path),
        params=params,
        files=files,
        data=data,
        headers=headers,
    )
    try:
        resp = requests.request(**request_kwargs)
        resp_json = resp.json()
    except requests.exceptions.ConnectionError:
        raise click.ClickException("Registry server is not responding.")
    except JSONDecodeError:
        resp_json = None

    if resp.status_code == 200:
        pass
    elif resp.status_code == 201:
        logger.debug("Successfully created!")
    elif resp.status_code == 403:
        raise click.ClickException(
            "You are not authenticated. " 'Please sign in with "aea login" command.'
        )
    elif resp.status_code == 500:
        raise click.ClickException("Registry internal server error.")
    elif resp.status_code == 404:
        raise click.ClickException("Not found in Registry.")
    elif resp.status_code == 409:
        raise click.ClickException(
            "Conflict in Registry. {}".format(resp_json["detail"])
        )
    elif resp.status_code == 400:
        if handle_400:
            raise click.ClickException(resp_json)
    else:
        raise click.ClickException(
            "Wrong server response. Status code: {}".format(resp.status_code)
        )
    if return_code:
        return resp_json, resp.status_code
    return resp_json


def download_file(url: str, cwd: str, timeout: float = FILE_DOWNLOAD_TIMEOUT) -> str:
    """
    Download file from URL and save it in CWD (current working directory).

    :param url: str url of the file to download.
    :param cwd: str path to current working directory.
    :param timeout: float. timeout to download a file

    :return: str path to downloaded file
    """
    local_filename = url.split("/")[-1]
    filepath = os.path.join(cwd, local_filename)
    # NOTE the stream=True parameter below
    response = requests.get(url, stream=True, timeout=timeout)
    if response.status_code == 200:
        with open(filepath, "wb") as f:
            f.write(response.raw.read())
    else:
        raise click.ClickException(
            "Wrong response from server when downloading package."
        )
    return filepath


def extract(source: str, target: str) -> None:
    """
    Extract tarball and remove source file.

    :param source: str path to a source tarball file.
    :param target: str path to target directory.

    :return: None
    """
    if source.endswith("tar.gz"):
        tar = tarfile.open(source, "r:gz")
        tar.extractall(path=target)
        tar.close()
    else:
        raise ValueError("Unknown file type: {}".format(source))

    os.remove(source)


def _rm_tarfiles() -> None:
    cwd = os.getcwd()
    for filename in os.listdir(cwd):
        if filename.endswith(".tar.gz"):
            filepath = os.path.join(cwd, filename)
            os.remove(filepath)


def clean_tarfiles(func: Callable) -> Callable:
    """Decorate func to clean tarfiles after executing."""

    def wrapper(*args: Any, **kwargs: Any) -> Callable:
        try:
            result = func(*args, **kwargs)
        except Exception as e:
            _rm_tarfiles()
            raise e
        else:
            _rm_tarfiles()
            return result

    return wrapper


def check_is_author_logged_in(author_name: str) -> None:
    """
    Check if current user's name equals to item's author.

    :param author_name: str item author username.

    :raise ClickException: if username and author's name are not equal.
    :return: None.
    """
    resp = cast(JSONLike, request_api("GET", "/rest-auth/user/", is_auth=True))
    if not author_name == resp["username"]:
        raise click.ClickException(
            "Author username is not equal to current logged in username "
            "(logged in: {}, author: {}). Please logout and then login correctly.".format(
                resp["username"], author_name
            )
        )


def is_auth_token_present() -> bool:
    """
    Check if any user is currently logged in.

    :return: bool is logged in.
    """
    return get_auth_token() is not None


def get_package_meta(
    obj_type: str, public_id: PublicId, aea_version: Optional[str] = None
) -> JSONLike:
    """
    Get package meta data from remote registry.

    Optionally filter by AEA version.

    :param obj_type: str. component type
    :param public_id: component public id
    :param aea_version: the AEA version (e.g. "0.1.0") or None.

    :return: dict with package details
    """
    params = dict(aea_version=aea_version) if aea_version else None
    api_path = f"/{obj_type}s/{public_id.author}/{public_id.name}/{public_id.version}"
    resp = cast(JSONLike, request_api("GET", api_path, params=params))
    return resp


def get_latest_public_id_mixed(
    ctx: Context,
    item_type: str,
    item_public_id: PublicId,
    aea_version: Optional[str] = None,
) -> PublicId:
    """
    Get latest public id of the message, mixed mode.

    That is, give priority to local registry, and fall back to remote registry
    in case of failure.

    :param ctx: the CLI context.
    :param item_type: the item type.
    :param item_public_id: the item public id.
    :param aea_version: the AEA version constraint, or None
    :return: the path to the found package.
    """
    try:
        _, item_config = find_item_locally(ctx, item_type, item_public_id)
        latest_item_public_id = item_config.public_id
    except click.ClickException:
        logger.debug(
            "Get latest public id from local registry failed, trying remote registry..."
        )
        # the following might raise exception, but we don't catch it this time
        package_meta = get_package_meta(
            item_type, item_public_id, aea_version=aea_version
        )
        latest_item_public_id = PublicId.from_str(cast(str, package_meta["public_id"]))
    return latest_item_public_id


def get_latest_version_available_in_registry(
    ctx: Context,
    item_type: str,
    item_public_id: PublicId,
    aea_version: Optional[str] = None,
) -> PublicId:
    """
    Get latest available package version public id.

    Optionally consider AEA version through the `aea_version` parameter.

    :param ctx: Context object.
    :param item_type: the item type.
    :param item_public_id: the item public id.
    :param aea_version: the AEA version (e.g. "0.1.0") or None.
    :return: the latest public id.
    """
    is_local = ctx.config.get("is_local")
    is_mixed = ctx.config.get("is_mixed")
    try:
        if is_mixed:
            latest_item_public_id = get_latest_public_id_mixed(
                ctx, item_type, item_public_id, aea_version
            )
        elif is_local:
            _, item_config = find_item_locally(ctx, item_type, item_public_id)
            latest_item_public_id = item_config.public_id
        else:
            package_meta = get_package_meta(item_type, item_public_id, aea_version)
            latest_item_public_id = PublicId.from_str(
                cast(str, package_meta["public_id"])
            )
    except Exception:  # pylint: disable=broad-except
        raise click.ClickException(
            f"Package {item_public_id} details can not be fetched from the registry!"
        )

    return latest_item_public_id



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the coin price skill."""

from pathlib import Path
from typing import cast

from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.protocols.prometheus.message import PrometheusMessage
from packages.fetchai.skills.coin_price.behaviours import CoinPriceBehaviour

from tests.conftest import ROOT_DIR


class TestSkillBehaviour(BaseSkillTestCase):
    """Test behaviours of coin price."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "coin_price")

    @classmethod
    def setup(cls, **kwargs):
        """Setup the test class."""
        super().setup()
        cls.coin_price_behaviour = cast(
            CoinPriceBehaviour, cls._skill.skill_context.behaviours.coin_price_behaviour
        )

    def test_send_http_request_message(self):
        """Test the send_http_request_message method of the coin_price behaviour."""
        self.coin_price_behaviour.send_http_request_message("GET", "some_url")
        self.assert_quantity_in_outbox(1)
        msg = cast(HttpMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=HttpMessage,
            performative=HttpMessage.Performative.REQUEST,
            url="some_url",
        )
        assert has_attributes, error_str

    def test_add_prometheus_metric(self):
        """Test the send_http_request_message method of the coin_price behaviour."""
        self.coin_price_behaviour.add_prometheus_metric(
            "some_metric", "Gauge", "some_description", {"label_key": "label_value"}
        )
        self.assert_quantity_in_outbox(1)
        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.ADD_METRIC,
            type="Gauge",
            title="some_metric",
            description="some_description",
            labels={"label_key": "label_value"},
        )
        assert has_attributes, error_str

    def test_update_prometheus_metric(self):
        """Test the test_update_prometheus_metric method of the coin_price behaviour."""
        self.coin_price_behaviour.update_prometheus_metric(
            "some_metric", "set", 0.0, {"label_key": "label_value"}
        )
        self.assert_quantity_in_outbox(1)
        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.UPDATE_METRIC,
            callable="set",
            title="some_metric",
            value=0.0,
            labels={"label_key": "label_value"},
        )
        assert has_attributes, error_str

    def test_setup(self):
        """Test that the setup method puts two messages (prometheus metrics) in the outbox by default."""
        self.coin_price_behaviour.setup()
        self.assert_quantity_in_outbox(2)

        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.ADD_METRIC,
            type="Gauge",
            title="num_retrievals",
        )
        assert has_attributes, error_str

        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.ADD_METRIC,
            type="Gauge",
            title="num_requests",
        )
        assert has_attributes, error_str

    def test_act(self):
        """Test that the act method of the coin_price behaviour puts one message (http request) in the outbox."""
        self.coin_price_behaviour.act()
        self.assert_quantity_in_outbox(1)

        url = self.coin_price_behaviour.context.coin_price_model.url
        coin_id = self.coin_price_behaviour.context.coin_price_model.coin_id
        currency = self.coin_price_behaviour.context.coin_price_model.currency

        query_url = url + f"simple/price?ids={coin_id}&vs_currencies={currency}"

        msg = cast(HttpMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=HttpMessage,
            performative=HttpMessage.Performative.REQUEST,
            url=query_url,
        )
        assert has_attributes, error_str

    def test_teardown(self):
        """Test that the teardown method of the coin_price behaviour leaves no messages in the outbox."""
        assert self.coin_price_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the strategy class of the simple_service_search skill."""

from pathlib import Path

from aea.helpers.search.models import Constraint, ConstraintType, Query
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.simple_service_search.strategy import Strategy

from tests.conftest import ROOT_DIR


class TestStrategy(BaseSkillTestCase):
    """Test Strategy of simple_service_search."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "simple_service_search"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.mocked_name_of_data = "some_name_for_data"
        cls.mocked_data_1 = (
            b'[{"type_1": "data_1", "type_2": "data_2", "type_3": "data_3"}]'
        )
        cls.mocked_data_2 = (
            b'{"type_1": "data_1", "type_2": "data_2", "type_3": "data_3"}'
        )
        cls.mocked_data_3 = b"some_non_jason_data"

        cls.search_query = {
            "search_key": "seller_service",
            "search_value": "generic_service",
            "constraint_type": "==",
        }
        cls.search_location = {"longitude": 0.1270, "latitude": 51.5194}
        cls.search_radius = 5.0
        cls.shared_state_key = "agents_found"

        cls.strategy = Strategy(
            search_query=cls.search_query,
            search_location=cls.search_location,
            search_radius=cls.search_radius,
            shared_storage_key=cls.shared_state_key,
            name="strategy",
            skill_context=cls._skill.skill_context,
        )

    def test_simple_properties(self):
        """Test simple properties of the Strategy class."""
        assert self.strategy.shared_storage_key == self.shared_state_key

    def test_get_query(self):
        """Test the get_query method of the Strategy class."""
        query = self.strategy.get_query()

        assert type(query) == Query
        assert len(query.constraints) == 2
        assert query.model is None

        location_constraint = Constraint(
            "location",
            ConstraintType(
                "distance", (self.strategy._agent_location, self.search_radius)
            ),
        )
        assert query.constraints[0] == location_constraint

        service_key_constraint = Constraint(
            self.search_query["search_key"],
            ConstraintType(
                self.search_query["constraint_type"], self.search_query["search_value"],
            ),
        )
        assert query.constraints[1] == service_key_constraint



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the dialogue classes of the simple_service_registration skill."""

from pathlib import Path
from typing import cast

from aea.helpers.search.models import Description
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.simple_service_registration.dialogues import (
    OefSearchDialogue,
    OefSearchDialogues,
)

from tests.conftest import ROOT_DIR


class TestDialogues(BaseSkillTestCase):
    """Test dialogue class of simple_service_registration."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "simple_service_registration"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.oef_search_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.mocked_description = Description({"foo1": 1, "bar1": 2})

    def test_oef_search_dialogues(self):
        """Test the OefSearchDialogue class."""
        _, dialogue = self.oef_search_dialogues.create(
            counterparty=self.skill.skill_context.search_service_address,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description=self.mocked_description,
        )
        assert dialogue.role == OefSearchDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the strategy class."""

import datetime
import json
import random
from typing import Any, Dict, List, Optional, Tuple, cast

from aea.helpers.search.models import Location

from packages.fetchai.connections.http_client.connection import (
    PUBLIC_ID as HTTP_CLIENT_PUBLIC_ID,
)
from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.skills.confirmation_aw3.dialogues import HttpDialogues
from packages.fetchai.skills.confirmation_aw3.registration_db import RegistrationDB
from packages.fetchai.skills.generic_buyer.strategy import GenericStrategy


class Strategy(GenericStrategy):
    """Strategy class extending Generic Strategy."""

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the strategy of the agent.

        :return: None
        """
        aw1_aea: Optional[str] = kwargs.pop("aw1_aea", None)
        if aw1_aea is None:
            raise ValueError("aw1_aea must be provided!")
        self.aw1_aea = aw1_aea
        self._locations = kwargs.pop("locations", {})
        if len(self._locations) == 0:
            raise ValueError("locations must have at least one entry")
        _, location = next(iter(self._locations.items()))
        kwargs["location"] = location
        self._search_queries = kwargs.pop("search_queries", {})
        if len(self._search_queries) == 0:
            raise ValueError("search_queries must have at least one entry")
        _, search_query = next(iter(self._search_queries.items()))
        kwargs["search_query"] = search_query
        kwargs["service_id"] = search_query["search_value"]
        leaderboard_url = kwargs.pop("leaderboard_url", None)
        if leaderboard_url is None:
            raise ValueError("No leader board url provided!")
        self.leaderboard_url = f"{leaderboard_url}/insert"
        leaderboard_token = kwargs.pop("leaderboard_token")
        if leaderboard_token is None:
            raise ValueError("No leader board token provided!")
        self.leaderboard_token = leaderboard_token
        super().__init__(**kwargs)

    def get_acceptable_counterparties(
        self, counterparties: Tuple[str, ...]
    ) -> Tuple[str, ...]:
        """
        Process counterparties and drop unacceptable ones.

        :return: list of counterparties
        """
        valid_counterparties: List[str] = []
        for counterparty in counterparties:
            if self.is_valid_counterparty(counterparty):
                valid_counterparties.append(counterparty)
        return tuple(valid_counterparties)

    def is_valid_counterparty(self, counterparty: str) -> bool:
        """
        Check if the counterparty is valid.

        :return: bool indicating validity
        """
        registration_db = cast(RegistrationDB, self.context.registration_db)
        if not registration_db.is_registered(counterparty):
            self.context.logger.info(
                f"Invalid counterparty={counterparty}, not registered!"
            )
            return False
        return True

    def successful_trade_with_counterparty(
        self, counterparty: str, data: Dict[str, str]
    ) -> None:
        """
        Do something on successful trade.

        :param counterparty: the counterparty address
        :param data: the data
        :return: False
        """
        registration_db = cast(RegistrationDB, self.context.registration_db)
        registration_db.set_trade(counterparty, datetime.datetime.now(), data)
        self.context.logger.info(f"Successful trade with={counterparty}.")
        developer_handle, nb_trades = registration_db.get_handle_and_trades(
            counterparty
        )
        http_dialogues = cast(HttpDialogues, self.context.http_dialogues)
        request_http_message, _ = http_dialogues.create(
            counterparty=str(HTTP_CLIENT_PUBLIC_ID),
            performative=HttpMessage.Performative.REQUEST,
            method="POST",
            url=self.leaderboard_url,
            headers="Content-Type: application/json; charset=utf-8",
            version="",
            body=json.dumps(
                {
                    "name": developer_handle,
                    "points": nb_trades,
                    "token": self.leaderboard_token,
                }
            ).encode("utf-8"),
        )
        self.context.outbox.put_message(message=request_http_message)
        self.context.logger.info(
            f"Notifying leaderboard: developer_handle={developer_handle}, nb_trades={nb_trades}."
        )

    def register_counterparty(self, counterparty: str, developer_handle: str) -> None:
        """
        Register a counterparty.

        :param counterparty: the counterparty address
        :param developer_handle: the developer handle
        :param data: the data
        :return: False
        """
        registration_db = cast(RegistrationDB, self.context.registration_db)
        registration_db.set_registered(counterparty, developer_handle)

    def update_search_query_params(self) -> None:
        """
        Update agent location and query for search.

        :return: None
        """
        search_query_type, search_query = random.choice(  # nosec
            list(self._search_queries.items())
        )
        self._search_query = search_query
        self._service_id = search_query["search_value"]
        location_name, location = random.choice(list(self._locations.items()))  # nosec
        self._agent_location = Location(
            latitude=location["latitude"], longitude=location["longitude"]
        )
        self.context.logger.info(
            f"New search_type={search_query_type} and location={location_name}."
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the sync connection module."""
import asyncio
import time
from unittest.mock import MagicMock, Mock, patch

import pytest

from aea.configurations.data_types import PublicId
from aea.connections.base import BaseSyncConnection
from aea.mail.base import Envelope


class SampleConnection(BaseSyncConnection):
    """Sample connection for testing."""

    MAX_WORKER_THREADS = 3

    connection_id = PublicId("test", "test", "0.1.0")
    PAUSE = 0.5

    def __init__(self, *args, **kwargs):
        """Init connection."""
        super().__init__(*args, **kwargs)
        self.main_called = False
        self.send_counter = 0
        self.on_connect_called = False
        self.on_disconnect_called = False

    def main(self):
        """Run main."""
        self.main_called = True
        envelope = Mock()
        envelope.message = "main"
        self.put_envelope(envelope)

    def on_send(self, envelope: Envelope) -> None:
        """Run on send."""
        time.sleep(self.PAUSE)
        resp_envelope = Mock()
        resp_envelope.message = f"resp for {str(envelope.message)}"
        self.put_envelope(resp_envelope)
        self.send_counter += 1

    def on_connect(self):
        """Run on connect."""
        self.on_connect_called = True

    def on_disconnect(self):
        """Run on disconnect."""
        self.on_disconnect_called = True


@pytest.mark.asyncio
async def test_sync_connection():
    """Test sync connection example."""
    conf = Mock()
    conf.public_id = SampleConnection.connection_id
    conf.config = {}
    con = SampleConnection(conf, MagicMock())
    await asyncio.wait_for(con.connect(), timeout=10)
    assert con.is_connected
    envelope = Mock()

    for i in range(10):
        envelope.message = str(i)
        await asyncio.wait_for(con.send(envelope), timeout=10)

    await asyncio.sleep(con.PAUSE * 1.5)
    await asyncio.wait_for(con.disconnect(), timeout=10)
    assert con.is_disconnected

    assert con.on_connect_called
    assert con.on_disconnect_called
    assert con.main_called
    assert con.send_counter == con.MAX_WORKER_THREADS

    with patch.object(con, "_ensure_connected"):
        envelope = await con.receive()
        assert envelope.message == "main"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the common types and interfaces used in the aea framework."""
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union


Address = str

Primitive = Union[str, int, bool, float]
_JSONDict = Dict[Any, Any]  # temporary placeholder
_JSONList = List[Any]  # temporary placeholder
_JSONType = Optional[Union[Primitive, _JSONDict, _JSONList]]
JSONLike = Dict[str, _JSONType]
PathLike = Union[os.PathLike, Path, str]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the tac negotiation skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import PropertyMock, patch

from aea.decision_maker.gop import GoalPursuitReadiness, OwnershipState, Preferences
from aea.helpers.search.models import Constraint, ConstraintType, Description, Query
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.tac_negotiation.behaviours import (
    GoodsRegisterAndSearchBehaviour,
    TransactionCleanUpBehaviour,
)
from packages.fetchai.skills.tac_negotiation.dialogues import (
    OefSearchDialogue,
    OefSearchDialogues,
)
from packages.fetchai.skills.tac_negotiation.strategy import Strategy
from packages.fetchai.skills.tac_negotiation.transactions import Transactions

from tests.conftest import ROOT_DIR


class TestSkillBehaviour(BaseSkillTestCase):
    """Test tac behaviour of tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        tac_dm_context_kwargs = {
            "goal_pursuit_readiness": GoalPursuitReadiness(),
            "ownership_state": OwnershipState(),
            "preferences": Preferences(),
        }
        super().setup(dm_context_kwargs=tac_dm_context_kwargs)
        cls.tac_negotiation = cast(
            GoodsRegisterAndSearchBehaviour,
            cls._skill.skill_context.behaviours.tac_negotiation,
        )
        cls.oef_search_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.goal_pursuit_readiness = (
            cls._skill.skill_context.decision_maker_handler_context.goal_pursuit_readiness
        )
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)
        cls.logger = cls._skill.skill_context.logger

        cls.mocked_description = Description({"foo1": 1, "bar1": 2})
        cls.mocked_query = Query(
            [Constraint("tac_service", ConstraintType("==", "both"))]
        )
        cls.sender = str(cls._skill.skill_context.skill_id)

    def test_init(self):
        """Test the __init__ method of the negotiation behaviour."""
        assert self.tac_negotiation.is_registered is False

    def test_setup(self):
        """Test the setup method of the negotiation behaviour."""
        assert self.tac_negotiation.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_act_i(self):
        """Test the act method of the negotiation behaviour where is_game_finished is True."""
        # setup
        self.skill.skill_context._agent_context._shared_state = {
            "is_game_finished": True
        }

        # operation
        self.tac_negotiation.act()

        # after
        self.assert_quantity_in_outbox(0)
        assert self.skill.skill_context.is_active is False

    def test_act_ii(self):
        """Test the act method of the negotiation behaviour where goal_pursuit_readiness is not ready."""
        # setup
        self.skill.skill_context._agent_context._shared_state = {
            "is_game_finished": False
        }
        self.goal_pursuit_readiness._status = GoalPursuitReadiness.Status.NOT_READY

        # operation
        self.tac_negotiation.act()

        # after
        self.assert_quantity_in_outbox(0)

    def test_act_iii(self):
        """Test the act method of the negotiation behaviour where is_registered is False."""
        # setup
        self.skill.skill_context._agent_context._shared_state = {
            "is_game_finished": False
        }
        self.goal_pursuit_readiness._status = GoalPursuitReadiness.Status.READY

        searching_for_types = [(True, "sellers"), (False, "buyers")]
        no_searches = len(searching_for_types)

        # operation
        with patch.object(
            self.strategy,
            "get_location_description",
            return_value=self.mocked_description,
        ):
            with patch.object(
                self.strategy,
                "get_register_service_description",
                return_value=self.mocked_description,
            ):
                with patch.object(
                    self.strategy,
                    "get_location_and_service_query",
                    return_value=self.mocked_query,
                ):
                    with patch.object(
                        type(self.strategy),
                        "searching_for_types",
                        new_callable=PropertyMock,
                        return_value=searching_for_types,
                    ):
                        with patch.object(self.logger, "log") as mock_logger:
                            self.tac_negotiation.act()

        # after
        self.assert_quantity_in_outbox(no_searches + 2)

        # _register_agent
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=self.sender,
            service_description=self.mocked_description,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "registering agent on SOEF.")

        # _register_service
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"updating service directory as {self.strategy.registering_as}.",
        )
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=self.sender,
            service_description=self.mocked_description,
        )
        assert has_attributes, error_str

        assert self.tac_negotiation.is_registered is True

        # _search_services
        for search in searching_for_types:
            message = self.get_message_from_outbox()
            has_attributes, error_str = self.message_has_attributes(
                actual_message=message,
                message_type=OefSearchMessage,
                performative=OefSearchMessage.Performative.SEARCH_SERVICES,
                to=self.skill.skill_context.search_service_address,
                sender=self.sender,
                query=self.mocked_query,
            )
            assert has_attributes, error_str

            assert (
                cast(
                    OefSearchDialogue, self.oef_search_dialogues.get_dialogue(message)
                ).is_seller_search
                == search[0]
            )

            mock_logger.assert_any_call(
                logging.INFO,
                f"searching for {search[1]}, search_id={message.dialogue_reference}.",
            )

    def test_act_iv(self):
        """Test the act method of the negotiation behaviour where is_registered is True."""
        # setup
        self.skill.skill_context._agent_context._shared_state = {
            "is_game_finished": False
        }
        self.goal_pursuit_readiness._status = GoalPursuitReadiness.Status.READY
        self.tac_negotiation.is_registered = True

        searching_for_types = [(True, "sellers"), (False, "buyers")]
        no_searches = len(searching_for_types)

        # operation
        with patch.object(
            self.strategy,
            "get_location_and_service_query",
            return_value=self.mocked_query,
        ):
            with patch.object(
                type(self.strategy),
                "searching_for_types",
                new_callable=PropertyMock,
                return_value=searching_for_types,
            ):
                with patch.object(self.logger, "log") as mock_logger:
                    self.tac_negotiation.act()

        # after
        self.assert_quantity_in_outbox(no_searches)

        # _search_services
        for search in searching_for_types:
            message = self.get_message_from_outbox()
            has_attributes, error_str = self.message_has_attributes(
                actual_message=message,
                message_type=OefSearchMessage,
                performative=OefSearchMessage.Performative.SEARCH_SERVICES,
                to=self.skill.skill_context.search_service_address,
                sender=self.sender,
                query=self.mocked_query,
            )
            assert has_attributes, error_str

            assert (
                cast(
                    OefSearchDialogue, self.oef_search_dialogues.get_dialogue(message)
                ).is_seller_search
                == search[0]
            )

            mock_logger.assert_any_call(
                logging.INFO,
                f"searching for {search[1]}, search_id={message.dialogue_reference}.",
            )

    def test_teardown_i(self):
        """Test the teardown method of the negotiation behaviour."""
        # setup
        self.tac_negotiation.is_registered = True

        # operation
        with patch.object(
            self.strategy,
            "get_unregister_service_description",
            return_value=self.mocked_description,
        ):
            with patch.object(
                self.strategy,
                "get_location_description",
                return_value=self.mocked_description,
            ):
                with patch.object(self.logger, "log") as mock_logger:
                    self.tac_negotiation.teardown()

        # after
        self.assert_quantity_in_outbox(2)

        # _unregister_service
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=self.sender,
            service_description=self.mocked_description,
        )
        assert has_attributes, error_str
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"unregistering from service directory as {self.strategy.registering_as}.",
        )

        # _unregister_agent
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=self.sender,
            service_description=self.mocked_description,
        )
        assert has_attributes, error_str
        mock_logger.assert_any_call(logging.INFO, "unregistering agent from SOEF.")

        # teardown
        assert self.tac_negotiation.is_registered is False

    def test_teardown_ii(self):
        """Test the teardown method of the negotiation behaviour where is_registered is False."""
        # setup
        self.tac_negotiation.is_registered = False

        # operation
        assert self.tac_negotiation.teardown() is None

        # after
        self.assert_quantity_in_outbox(0)


class TestTransactionCleanUpBehaviour(BaseSkillTestCase):
    """Test clean_up behaviour of tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.clean_up_behaviour = cast(
            TransactionCleanUpBehaviour, cls._skill.skill_context.behaviours.clean_up
        )
        cls.transactions = cast(Transactions, cls._skill.skill_context.transactions)

    def test_setup(self):
        """Test the setup method of the clean_up behaviour."""
        assert self.clean_up_behaviour.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_act(self):
        """Test the act method of the clean_up behaviour."""
        # operation
        with patch.object(
            self.transactions, "update_confirmed_transactions"
        ) as mocked_update:
            with patch.object(
                self.transactions, "cleanup_pending_transactions"
            ) as mocked_cleanup:
                self.clean_up_behaviour.act()

        # after
        self.assert_quantity_in_outbox(0)
        mocked_update.assert_called_once()
        mocked_cleanup.assert_called_once()

    def test_teardown(self):
        """Test the teardown method of the clean_up behaviour."""
        assert self.clean_up_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the confirmation aw2 skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.crypto.ledger_apis import LedgerApis
from aea.protocols.dialogue.base import DialogueMessage

from packages.fetchai.protocols.default.dialogues import DefaultDialogue
from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.register.message import RegisterMessage
from packages.fetchai.skills.confirmation_aw2.dialogues import DefaultDialogues
from packages.fetchai.skills.confirmation_aw2.handlers import DefaultHandler
from packages.fetchai.skills.confirmation_aw2.strategy import Strategy

from tests.conftest import ROOT_DIR
from tests.test_packages.test_skills.test_confirmation_aw2.intermediate_class import (
    ConfirmationAW2TestCase,
)


class TestDefaultHandler(ConfirmationAW2TestCase):
    """Test default handler of confirmation aw2."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "confirmation_aw2")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.default_handler = cast(
            DefaultHandler, cls._skill.skill_context.handlers.default_handler
        )
        cls.logger = cls._skill.skill_context.logger
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)

        cls.default_dialogues = cast(
            DefaultDialogues, cls._skill.skill_context.default_dialogues
        )

        cls.list_of_default_messages = (
            DialogueMessage(
                DefaultMessage.Performative.BYTES, {"content": b"some_content"}
            ),
        )

        cls.confirmed_aea = b"ConfirmedAEA"
        cls.developer_handle = b"DeveloperHandle"

    def test_setup(self):
        """Test the setup method of the default handler."""
        assert self.default_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the register handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=DefaultMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=DefaultMessage.Performative.BYTES,
            content=b"some_content",
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.default_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid default message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_bytes_i(self):
        """Test the _handle_bytes method of the default handler where the sender IS aw1_aea."""
        # setup
        incoming_message = cast(
            RegisterMessage,
            self.build_incoming_message(
                message_type=DefaultMessage,
                performative=DefaultMessage.Performative.BYTES,
                content=self.confirmed_aea + b"_" + self.developer_handle,
                sender=self.aw1_aea,
            ),
        )

        # operation
        with patch.object(LedgerApis, "is_valid_address", return_value=True):
            with patch.object(self.logger, "log") as mock_logger:
                with patch.object(
                    self.strategy, "register_counterparty"
                ) as mock_register:
                    self.default_handler.handle(incoming_message)

        # after
        mock_register.called_once()

        mock_logger.assert_any_call(
            logging.INFO,
            f"adding confirmed_aea={self.confirmed_aea.decode('utf-8')} with developer_handle={self.developer_handle.decode('utf-8')} to db.",
        )

    def test_handle_bytes_ii(self):
        """Test the _handle_bytes method of the default handler where the content is undecodable."""
        # setup
        incorrect_content = "some_incorrect_content"

        incoming_message = cast(
            RegisterMessage,
            self.build_incoming_message(
                message_type=DefaultMessage,
                performative=DefaultMessage.Performative.BYTES,
                content=incorrect_content,
                sender=self.aw1_aea,
            ),
        )

        # operation
        with patch.object(LedgerApis, "is_valid_address", return_value=True):
            with patch.object(self.logger, "log") as mock_logger:
                with patch.object(
                    self.strategy, "register_counterparty"
                ) as mock_register:
                    self.default_handler.handle(incoming_message)

        # after
        mock_register.called_once()

        mock_logger.assert_any_call(
            logging.WARNING, "received invalid developer_handle=."
        )

    def test_handle_bytes_iii(self):
        """Test the _handle_bytes method of the default handler where is_valid_address is False."""
        # setup
        incoming_message = cast(
            RegisterMessage,
            self.build_incoming_message(
                message_type=DefaultMessage,
                performative=DefaultMessage.Performative.BYTES,
                content=self.confirmed_aea + b"_" + self.developer_handle,
                sender=self.aw1_aea,
            ),
        )

        # operation
        with patch.object(LedgerApis, "is_valid_address", return_value=False):
            with patch.object(self.logger, "log") as mock_logger:
                with patch.object(
                    self.strategy, "register_counterparty"
                ) as mock_register:
                    self.default_handler.handle(incoming_message)

        # after
        mock_register.called_once()

        default_dialogue = cast(
            DefaultDialogue, self.default_dialogues.get_dialogue(incoming_message)
        )

        mock_logger.assert_any_call(
            logging.WARNING,
            f"received invalid address={self.confirmed_aea.decode('utf-8')} in dialogue={default_dialogue}.",
        )

    def test_handle_bytes_iv(self):
        """Test the _handle_bytes method of the default handler where the sender is NOT aw1_aea."""
        # setup
        incoming_message = cast(
            RegisterMessage,
            self.build_incoming_message(
                message_type=DefaultMessage,
                performative=DefaultMessage.Performative.BYTES,
                content=self.confirmed_aea + b"_" + self.developer_handle,
                sender="some_other_aea",
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(self.strategy, "register_counterparty") as mock_register:
                self.default_handler.handle(incoming_message)

        # after
        mock_register.called_once()

        default_dialogue = cast(
            DefaultDialogue, self.default_dialogues.get_dialogue(incoming_message)
        )

        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle default message of performative={incoming_message.performative} in dialogue={default_dialogue}. Invalid sender={incoming_message.sender}",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the default handler."""
        # setup
        default_dialogue = cast(
            DefaultDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.default_dialogues,
                messages=self.list_of_default_messages[:1],
            ),
        )
        incoming_message = cast(
            RegisterMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=default_dialogue,
                performative=DefaultMessage.Performative.ERROR,
                error_code=DefaultMessage.ErrorCode.DECODING_ERROR,
                error_msg="some_error_message",
                error_data={"some_key": b"some_value"},
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.default_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle default message of performative={incoming_message.performative} in dialogue={default_dialogue}.",
        )

    def test_teardown(self):
        """Test the teardown method of the default handler."""
        assert self.default_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a simple Fetch oracle client behaviour."""

from typing import Any, cast

from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.connections.ledger.base import CONNECTION_ID as LEDGER_API_ADDRESS
from packages.fetchai.contracts.fet_erc20.contract import (
    PUBLIC_ID as FET_ERC20_PUBLIC_ID,
)
from packages.fetchai.contracts.oracle_client.contract import (
    PUBLIC_ID as CLIENT_CONTRACT_PUBLIC_ID,
)
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.skills.simple_oracle_client.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
)
from packages.fetchai.skills.simple_oracle_client.strategy import Strategy


DEFAULT_QUERY_INTERVAL = 5


class SimpleOracleClientBehaviour(TickerBehaviour):
    """This class implements a behaviour that deploys a Fetch oracle client contract."""

    def __init__(self, **kwargs: Any):
        """Initialise the behaviour."""
        query_interval = kwargs.pop(
            "query_interval", DEFAULT_QUERY_INTERVAL
        )  # type: int
        super().__init__(tick_interval=query_interval, **kwargs)

    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """

        self.context.logger.info("Setting up Fetch oracle client contract...")
        strategy = cast(Strategy, self.context.strategy)

        if not strategy.is_client_contract_deployed:
            self._request_contract_deploy_transaction()
        else:
            self.context.logger.info(
                "Fetch oracle client contract address already added"
            )

    def act(self) -> None:
        """
        Implement the act.

        :return: None
        """

        strategy = cast(Strategy, self.context.strategy)

        if not strategy.is_client_contract_deployed:
            self.context.logger.info("Oracle client contract not yet deployed")
            return
        if not strategy.is_oracle_transaction_approved:
            self.context.logger.info(
                "Oracle client contract not yet approved to spend tokens"
            )
            self._request_approve_transaction()
            return

        # Call contract function that queries oracle value
        self.context.logger.info("Calling contract to request oracle value...")
        self._request_query_transaction()

    def _request_contract_deploy_transaction(self) -> None:
        """
        Request contract deployment transaction

        :return: None
        """
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=str(LEDGER_API_ADDRESS),
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=str(CLIENT_CONTRACT_PUBLIC_ID),
            callable="get_deploy_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {
                    "deployer_address": self.context.agent_address,
                    "fetchOracleContractAddress": strategy.oracle_contract_address,
                    "gas": strategy.default_gas_deploy,
                }
            ),
        )
        contract_api_dialogue = cast(ContractApiDialogue, contract_api_dialogue,)
        contract_api_dialogue.terms = strategy.get_deploy_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting contract deployment transaction...")

    def _request_approve_transaction(self) -> None:
        """
        Request transaction that approves client contract to spend tokens on behalf of sender

        :return: None
        """
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=str(LEDGER_API_ADDRESS),
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=str(FET_ERC20_PUBLIC_ID),
            contract_address=strategy.erc20_address,
            callable="get_approve_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {
                    "from_address": self.context.agent_address,
                    "spender": strategy.client_contract_address,
                    "amount": strategy.approve_amount,
                    "gas": strategy.default_gas_approve,
                }
            ),
        )
        contract_api_dialogue = cast(ContractApiDialogue, contract_api_dialogue)
        contract_api_dialogue.terms = strategy.get_approve_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting query transaction...")

    def _request_query_transaction(self) -> None:
        """
        Request transaction that requests value from Fetch oracle contract

        :return: None
        """
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=str(LEDGER_API_ADDRESS),
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=str(CLIENT_CONTRACT_PUBLIC_ID),
            contract_address=strategy.client_contract_address,
            callable="get_query_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {
                    "from_address": self.context.agent_address,
                    "query_function": strategy.query_function,
                    "gas": strategy.default_gas_query,
                }
            ),
        )
        contract_api_dialogue = cast(ContractApiDialogue, contract_api_dialogue)
        contract_api_dialogue.terms = strategy.get_query_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting query transaction...")

    def teardown(self) -> None:
        """
        Implement the task teardown.

        :return: None
        """



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""This module contains the tests of the ledger API connection module."""
import asyncio
import logging
from typing import cast
from unittest.mock import Mock, patch

import pytest
from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_fetchai import FetchAICrypto

from aea.common import Address
from aea.configurations.base import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.crypto.ledger_apis import LedgerApis
from aea.crypto.registries import make_crypto, make_ledger_api
from aea.helpers.async_utils import AsyncState
from aea.helpers.transaction.base import (
    RawTransaction,
    SignedTransaction,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from aea.mail.base import Envelope, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue

from packages.fetchai.connections.ledger.connection import LedgerConnection
from packages.fetchai.connections.ledger.ledger_dispatcher import (
    LedgerApiRequestDispatcher,
)
from packages.fetchai.protocols.ledger_api.custom_types import Kwargs
from packages.fetchai.protocols.ledger_api.dialogues import LedgerApiDialogue
from packages.fetchai.protocols.ledger_api.dialogues import (
    LedgerApiDialogues as BaseLedgerApiDialogues,
)
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage

from tests.conftest import (
    ETHEREUM_PRIVATE_KEY_PATH,
    FETCHAI_ADDRESS_ONE,
    FETCHAI_TESTNET_CONFIG,
)


logger = logging.getLogger(__name__)


ledger_ids = pytest.mark.parametrize(
    "ledger_id,address",
    [
        (FetchAICrypto.identifier, FETCHAI_ADDRESS_ONE),
        (EthereumCrypto.identifier, EthereumCrypto(ETHEREUM_PRIVATE_KEY_PATH).address),
    ],
)

SOME_SKILL_ID = "some/skill:0.1.0"


class LedgerApiDialogues(BaseLedgerApiDialogues):
    """The dialogues class keeps track of all ledger_api dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return LedgerApiDialogue.Role.AGENT

        BaseLedgerApiDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
        )


@pytest.mark.integration
@pytest.mark.ledger
@pytest.mark.asyncio
@ledger_ids
async def test_get_balance(
    ledger_id,
    address,
    ledger_apis_connection: Connection,
    update_default_ethereum_ledger_api,
    ethereum_testnet_config,
    ganache,
):
    """Test get balance."""
    import aea  # noqa # to load registries

    if ledger_id == FetchAICrypto.identifier:
        config = FETCHAI_TESTNET_CONFIG
    else:
        config = ethereum_testnet_config

    ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)
    request, ledger_api_dialogue = ledger_api_dialogues.create(
        counterparty=str(ledger_apis_connection.connection_id),
        performative=LedgerApiMessage.Performative.GET_BALANCE,
        ledger_id=ledger_id,
        address=address,
    )
    envelope = Envelope(to=request.to, sender=request.sender, message=request,)

    await ledger_apis_connection.send(envelope)
    await asyncio.sleep(0.01)
    response = await ledger_apis_connection.receive()

    assert response is not None
    assert type(response.message) == LedgerApiMessage
    response_msg = cast(LedgerApiMessage, response.message)
    response_dialogue = ledger_api_dialogues.update(response_msg)
    assert response_dialogue == ledger_api_dialogue
    assert response_msg.performative == LedgerApiMessage.Performative.BALANCE
    actual_balance_amount = response_msg.balance
    expected_balance_amount = make_ledger_api(ledger_id, **config).get_balance(address)
    assert actual_balance_amount == expected_balance_amount


@pytest.mark.integration
@pytest.mark.ledger
@pytest.mark.asyncio
@ledger_ids
async def test_get_state(
    ledger_id,
    address,
    ledger_apis_connection: Connection,
    update_default_ethereum_ledger_api,
    ethereum_testnet_config,
    ganache,
):
    """Test get state."""
    import aea  # noqa # to load registries

    if ledger_id == FetchAICrypto.identifier:
        config = FETCHAI_TESTNET_CONFIG
    else:
        config = ethereum_testnet_config

    if "ethereum" in ledger_id:
        callable_name = "getBlock"
    else:
        callable_name = "blocks"
    args = ("latest",)
    kwargs = Kwargs({})

    ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)
    request, ledger_api_dialogue = ledger_api_dialogues.create(
        counterparty=str(ledger_apis_connection.connection_id),
        performative=LedgerApiMessage.Performative.GET_STATE,
        ledger_id=ledger_id,
        callable=callable_name,
        args=args,
        kwargs=kwargs,
    )
    envelope = Envelope(to=request.to, sender=request.sender, message=request,)

    await ledger_apis_connection.send(envelope)
    await asyncio.sleep(0.01)
    response = await ledger_apis_connection.receive()

    assert response is not None
    assert type(response.message) == LedgerApiMessage
    response_msg = cast(LedgerApiMessage, response.message)
    response_dialogue = ledger_api_dialogues.update(response_msg)
    assert response_dialogue == ledger_api_dialogue

    assert (
        response_msg.performative == LedgerApiMessage.Performative.STATE
    ), response_msg
    actual_block = response_msg.state.body
    expected_block = make_ledger_api(ledger_id, **config).get_state(
        callable_name, *args
    )
    assert actual_block == expected_block


@pytest.mark.integration
@pytest.mark.ledger
@pytest.mark.asyncio
async def test_send_signed_transaction_ethereum(
    ledger_apis_connection: Connection, update_default_ethereum_ledger_api, ganache
):
    """Test send signed transaction with Ethereum APIs."""
    import aea  # noqa # to load registries

    crypto1 = make_crypto(
        EthereumCrypto.identifier, private_key_path=ETHEREUM_PRIVATE_KEY_PATH
    )
    crypto2 = make_crypto(EthereumCrypto.identifier)
    ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)

    amount = 40000
    fee = 30000

    request, ledger_api_dialogue = ledger_api_dialogues.create(
        counterparty=str(ledger_apis_connection.connection_id),
        performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
        terms=Terms(
            ledger_id=EthereumCrypto.identifier,
            sender_address=crypto1.address,
            counterparty_address=crypto2.address,
            amount_by_currency_id={"ETH": -amount},
            quantities_by_good_id={"some_service_id": 1},
            is_sender_payable_tx_fee=True,
            nonce="",
            fee_by_currency_id={"ETH": fee},
            chain_id=3,
        ),
    )
    request = cast(LedgerApiMessage, request)
    envelope = Envelope(to=request.to, sender=request.sender, message=request,)
    await ledger_apis_connection.send(envelope)
    await asyncio.sleep(0.01)
    response = await ledger_apis_connection.receive()

    assert response is not None
    assert type(response.message) == LedgerApiMessage
    response_message = cast(LedgerApiMessage, response.message)
    assert (
        response_message.performative == LedgerApiMessage.Performative.RAW_TRANSACTION
    )
    response_dialogue = ledger_api_dialogues.update(response_message)
    assert response_dialogue == ledger_api_dialogue
    assert type(response_message.raw_transaction) == RawTransaction
    assert response_message.raw_transaction.ledger_id == request.terms.ledger_id

    signed_transaction = crypto1.sign_transaction(response_message.raw_transaction.body)
    request = cast(
        LedgerApiMessage,
        ledger_api_dialogue.reply(
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            target_message=response_message,
            signed_transaction=SignedTransaction(
                EthereumCrypto.identifier, signed_transaction
            ),
        ),
    )
    envelope = Envelope(to=request.to, sender=request.sender, message=request,)
    await ledger_apis_connection.send(envelope)
    await asyncio.sleep(0.01)
    response = await ledger_apis_connection.receive()

    assert response is not None
    assert type(response.message) == LedgerApiMessage
    response_message = cast(LedgerApiMessage, response.message)
    assert (
        response_message.performative != LedgerApiMessage.Performative.ERROR
    ), f"Received error: {response_message.message}"
    assert (
        response_message.performative
        == LedgerApiMessage.Performative.TRANSACTION_DIGEST
    )
    response_dialogue = ledger_api_dialogues.update(response_message)
    assert response_dialogue == ledger_api_dialogue
    assert type(response_message.transaction_digest) == TransactionDigest
    assert type(response_message.transaction_digest.body) == str
    assert (
        response_message.transaction_digest.ledger_id
        == request.signed_transaction.ledger_id
    )
    assert type(response_message.transaction_digest.body.startswith("0x"))

    request = cast(
        LedgerApiMessage,
        ledger_api_dialogue.reply(
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
            target_message=response_message,
            transaction_digest=response_message.transaction_digest,
        ),
    )
    envelope = Envelope(to=request.to, sender=request.sender, message=request,)
    await ledger_apis_connection.send(envelope)
    await asyncio.sleep(0.01)
    response = await ledger_apis_connection.receive()

    assert response is not None
    assert type(response.message) == LedgerApiMessage
    response_message = cast(LedgerApiMessage, response.message)
    assert (
        response_message.performative
        == LedgerApiMessage.Performative.TRANSACTION_RECEIPT
    )
    response_dialogue = ledger_api_dialogues.update(response_message)
    assert response_dialogue == ledger_api_dialogue
    assert type(response_message.transaction_receipt) == TransactionReceipt
    assert response_message.transaction_receipt.receipt is not None
    assert response_message.transaction_receipt.transaction is not None
    assert (
        response_message.transaction_receipt.ledger_id
        == request.transaction_digest.ledger_id
    )
    assert LedgerApis.is_transaction_settled(
        response_message.transaction_receipt.ledger_id,
        response_message.transaction_receipt.receipt,
    ), "Transaction not settled."


@pytest.mark.asyncio
async def test_unsupported_protocol(ledger_apis_connection: LedgerConnection):
    """Test fail on protocol not supported."""
    envelope = Envelope(
        to=str(ledger_apis_connection.connection_id),
        sender="test/skill:0.1.0",
        protocol_specification_id=PublicId.from_str("author/package_name:0.1.0"),
        message=b"message",
    )
    with pytest.raises(ValueError):
        ledger_apis_connection._schedule_request(envelope)


@pytest.mark.asyncio
async def test_new_message_wait_flag(ledger_apis_connection: LedgerConnection):
    """Test wait for new message."""
    task = asyncio.ensure_future(ledger_apis_connection.receive())
    await asyncio.sleep(0.1)
    assert not task.done()
    task.cancel()


@pytest.mark.asyncio
async def test_no_balance():
    """Test no balance."""
    dispatcher = LedgerApiRequestDispatcher(AsyncState())
    mock_api = Mock()
    message = LedgerApiMessage(
        performative=LedgerApiMessage.Performative.GET_BALANCE,
        dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
        ledger_id=EthereumCrypto.identifier,
        address="test",
    )
    message.to = dispatcher.dialogues.self_address
    message.sender = "test"
    dialogue = dispatcher.dialogues.update(message)
    assert dialogue is not None
    mock_api.get_balance.return_value = None
    msg = dispatcher.get_balance(mock_api, message, dialogue)

    assert msg.performative == LedgerApiMessage.Performative.ERROR


@pytest.mark.asyncio
async def test_no_raw_tx():
    """Test no raw tx returned."""
    dispatcher = LedgerApiRequestDispatcher(AsyncState())
    mock_api = Mock()
    message = LedgerApiMessage(
        performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
        dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
        terms=Terms(
            ledger_id=EthereumCrypto.identifier,
            sender_address="1111",
            counterparty_address="22222",
            amount_by_currency_id={"ETH": -1},
            quantities_by_good_id={"some_service_id": 1},
            is_sender_payable_tx_fee=True,
            nonce="",
            fee_by_currency_id={"ETH": 10},
            chain_id=3,
        ),
    )
    message.to = dispatcher.dialogues.self_address
    message.sender = "test"
    dialogue = dispatcher.dialogues.update(message)
    assert dialogue is not None
    mock_api.get_transfer_transaction.return_value = None
    msg = dispatcher.get_raw_transaction(mock_api, message, dialogue)

    assert msg.performative == LedgerApiMessage.Performative.ERROR


@pytest.mark.asyncio
async def test_attempts_get_transaction_receipt():
    """Test retry and sleep."""
    dispatcher = LedgerApiRequestDispatcher(AsyncState(ConnectionStates.connected))
    mock_api = Mock()
    message = LedgerApiMessage(
        performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
        dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
        transaction_digest=TransactionDigest("asdad", "sdfdsf"),
    )
    message.to = dispatcher.dialogues.self_address
    message.sender = "test"
    dialogue = dispatcher.dialogues.update(message)
    assert dialogue is not None
    mock_api.get_transaction.return_value = None
    mock_api.is_transaction_settled.return_value = True
    with patch.object(dispatcher, "MAX_ATTEMPTS", 2):
        with patch.object(dispatcher, "TIMEOUT", 0.001):
            msg = dispatcher.get_transaction_receipt(mock_api, message, dialogue)

    assert msg.performative == LedgerApiMessage.Performative.ERROR



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the `aea scaffold protocol` sub-command."""
import filecmp
import json
import os
import shutil
import tempfile
import unittest.mock
from pathlib import Path
from unittest.mock import patch

import jsonschema
import yaml
from jsonschema import Draft4Validator, ValidationError

from aea import AEA_DIR
from aea.cli import cli
from aea.configurations.base import DEFAULT_PROTOCOL_CONFIG_FILE
from aea.configurations.loader import make_jsonschema_base_uri

from tests.conftest import (
    AUTHOR,
    CLI_LOG_OPTION,
    CONFIGURATION_SCHEMA_DIR,
    CliRunner,
    PROTOCOL_CONFIGURATION_SCHEMA,
    ROOT_DIR,
)


class TestScaffoldProtocol:
    """Test that the command 'aea scaffold protocol' works correctly in correct preconditions."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.resource_name = "myresource"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        dir_path = Path("packages")
        tmp_dir = cls.t / dir_path
        src_dir = cls.cwd / Path(ROOT_DIR, dir_path)
        shutil.copytree(str(src_dir), str(tmp_dir))
        cls.schema = json.load(open(PROTOCOL_CONFIGURATION_SCHEMA))
        cls.resolver = jsonschema.RefResolver(
            make_jsonschema_base_uri(Path(CONFIGURATION_SCHEMA_DIR).absolute()),
            cls.schema,
        )
        cls.validator = Draft4Validator(cls.schema, resolver=cls.resolver)

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)
        # scaffold protocol
        with patch("click.confirm", return_value=True) as confirm_mock:
            cls.result = cls.runner.invoke(
                cli,
                [*CLI_LOG_OPTION, "scaffold", "protocol", cls.resource_name],
                standalone_mode=False,
            )
            confirm_mock.assert_called_once_with(
                "We highly recommend auto-generating protocols with the aea generate command. Do you really want to continue scaffolding?"
            )

    def test_exit_code_equal_to_0(self):
        """Test that the exit code is equal to 0."""
        assert self.result.exit_code == 0

    def test_resource_folder_contains_module_message(self):
        """Test that the resource folder contains scaffold message.py module."""
        p = Path(self.t, self.agent_name, "protocols", self.resource_name, "message.py")
        original = Path(AEA_DIR, "protocols", "scaffold", "message.py")
        assert filecmp.cmp(p, original)

    def test_resource_folder_contains_module_protocol(self):
        """Test that the resource folder contains scaffold protocol.py module."""
        p = Path(
            self.t, self.agent_name, "protocols", self.resource_name, "serialization.py"
        )
        original = Path(AEA_DIR, "protocols", "scaffold", "serialization.py")
        assert filecmp.cmp(p, original)

    def test_resource_folder_contains_configuration_file(self):
        """Test that the resource folder contains a good configuration file."""
        p = Path(
            self.t,
            self.agent_name,
            "protocols",
            self.resource_name,
            DEFAULT_PROTOCOL_CONFIG_FILE,
        )
        config_file = yaml.safe_load(open(p))
        self.validator.validate(instance=config_file)

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestScaffoldProtocolFailsWhenDirectoryAlreadyExists:
    """Test that the command 'aea scaffold protocol' fails when a folder with 'scaffold' name already."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.resource_name = "myresource"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        dir_path = Path("packages")
        tmp_dir = cls.t / dir_path
        src_dir = cls.cwd / Path(ROOT_DIR, dir_path)
        shutil.copytree(str(src_dir), str(tmp_dir))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)
        # create a dummy 'myresource' folder
        Path(cls.t, cls.agent_name, "protocols", cls.resource_name).mkdir(
            exist_ok=False, parents=True
        )
        with patch("click.confirm", return_value=True):
            cls.result = cls.runner.invoke(
                cli,
                [*CLI_LOG_OPTION, "scaffold", "protocol", cls.resource_name],
                standalone_mode=False,
            )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_error_message_protocol_already_existing(self):
        """Test that the log error message is fixed.

        The expected message is: 'A protocol with name '{protocol_name}' already exists. Aborting...'
        """
        s = "A protocol with this name already exists. Please choose a different name and try again."
        assert self.result.exception.message == s

    def test_resource_directory_exists(self):
        """Test that the resource directory still exists.

        This means that after every failure, we make sure we restore the previous state.
        """
        assert Path(self.t, self.agent_name, "protocols", self.resource_name).exists()

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestScaffoldProtocolFailsWhenProtocolAlreadyExists:
    """Test that the command 'aea add protocol' fails when the protocol already exists."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.resource_name = "myresource"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        dir_path = Path("packages")
        tmp_dir = cls.t / dir_path
        src_dir = cls.cwd / Path(ROOT_DIR, dir_path)
        shutil.copytree(str(src_dir), str(tmp_dir))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0
        os.chdir(cls.agent_name)
        # add protocol first time
        with patch("click.confirm", return_value=True):
            result = cls.runner.invoke(
                cli,
                [*CLI_LOG_OPTION, "scaffold", "protocol", cls.resource_name],
                standalone_mode=False,
            )
            assert result.exit_code == 0
            # scaffold protocol with the same protocol name

            cls.result = cls.runner.invoke(
                cli,
                [*CLI_LOG_OPTION, "scaffold", "protocol", cls.resource_name],
                standalone_mode=False,
            )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_error_message_protocol_already_existing(self):
        """Test that the log error message is fixed.

        The expected message is: 'A protocol with name '{protocol_name}' already exists. Aborting...'
        """
        s = "A protocol with name '{}' already exists. Aborting...".format(
            self.resource_name
        )
        assert s in self.result.exception.message

    def test_resource_directory_exists(self):
        """Test that the resource directory still exists.

        This means that after every failure, we make sure we restore the previous state.
        """
        assert Path(self.t, self.agent_name, "protocols", self.resource_name).exists()

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestScaffoldProtocolFailsWhenConfigFileIsNotCompliant:
    """Test that the command 'aea scaffold protocol' fails when the configuration file is not compliant with the schema."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.resource_name = "myresource"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        dir_path = Path("packages")
        tmp_dir = cls.t / dir_path
        src_dir = cls.cwd / Path(ROOT_DIR, dir_path)
        shutil.copytree(str(src_dir), str(tmp_dir))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0

        # change the dumping of yaml module to raise an exception.
        cls.patch = unittest.mock.patch(
            "yaml.dump", side_effect=ValidationError("test error message")
        )
        cls.patch.start()

        os.chdir(cls.agent_name)
        with patch("click.confirm", return_value=True):
            cls.result = cls.runner.invoke(
                cli,
                [*CLI_LOG_OPTION, "scaffold", "protocol", cls.resource_name],
                standalone_mode=False,
            )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_configuration_file_not_valid(self):
        """Test that the log error message is fixed.

        The expected message is: 'Cannot find protocol: '{protocol_name}'
        """
        s = "Error when validating the protocol configuration file."
        assert self.result.exception.message == s

    def test_resource_directory_does_not_exists(self):
        """Test that the resource directory does not exist.

        This means that after every failure, we make sure we restore the previous state.
        """
        assert not Path(
            self.t, self.agent_name, "protocols", self.resource_name
        ).exists()

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        cls.patch.stop()
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestScaffoldProtocolFailsWhenExceptionOccurs:
    """Test that the command 'aea scaffold protocol' fails when the configuration file is not compliant with the schema."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.runner = CliRunner()
        cls.agent_name = "myagent"
        cls.resource_name = "myresource"
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        dir_path = Path("packages")
        tmp_dir = cls.t / dir_path
        src_dir = cls.cwd / Path(ROOT_DIR, dir_path)
        shutil.copytree(str(src_dir), str(tmp_dir))

        os.chdir(cls.t)
        result = cls.runner.invoke(
            cli, [*CLI_LOG_OPTION, "init", "--local", "--author", AUTHOR]
        )
        assert result.exit_code == 0
        result = cls.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "create", "--local", cls.agent_name],
            standalone_mode=False,
        )
        assert result.exit_code == 0

        cls.patch = unittest.mock.patch(
            "shutil.copytree", side_effect=Exception("unknwon exception")
        )
        cls.patch.start()

        os.chdir(cls.agent_name)
        with patch("click.confirm", return_value=True):
            cls.result = cls.runner.invoke(
                cli,
                [*CLI_LOG_OPTION, "scaffold", "protocol", cls.resource_name],
                standalone_mode=False,
            )

    def test_exit_code_equal_to_1(self):
        """Test that the exit code is equal to 1 (i.e. catchall for general errors)."""
        assert self.result.exit_code == 1

    def test_resource_directory_does_not_exists(self):
        """Test that the resource directory does not exist.

        This means that after every failure, we make sure we restore the previous state.
        """
        assert not Path(
            self.t, self.agent_name, "protocols", self.resource_name
        ).exists()

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        cls.patch.stop()
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the simple_service_registration skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.helpers.search.models import Description
from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.simple_service_registration.dialogues import (
    OefSearchDialogues,
)
from packages.fetchai.skills.simple_service_registration.handlers import (
    OefSearchHandler,
)

from tests.conftest import ROOT_DIR


class TestOefSearchHandler(BaseSkillTestCase):
    """Test oef_search handler of simple_service_registration."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "simple_service_registration"
    )
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.oef_search_handler = cast(
            OefSearchHandler, cls._skill.skill_context.handlers.oef_search
        )
        cls.logger = cls._skill.skill_context.logger

        cls.oef_search_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )

        cls.data = b"some_body"
        cls.mocked_description = Description({"foo1": 1, "bar1": 2})

        cls.list_of_messages = (
            DialogueMessage(
                OefSearchMessage.Performative.REGISTER_SERVICE,
                {"service_description": cls.mocked_description},
            ),
        )

    def test_setup(self):
        """Test the setup method of the oef_search handler."""
        assert self.oef_search_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the oef_search handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=OefSearchMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=OefSearchMessage.Performative.OEF_ERROR,
            oef_error_operation=OefSearchMessage.OefErrorOperation.REGISTER_SERVICE,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid oef_search message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_error(self):
        """Test the _handle_error method of the oef_search handler."""
        # setup
        oef_search_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_search_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=oef_search_dialogue,
            performative=OefSearchMessage.Performative.OEF_ERROR,
            oef_error_operation=OefSearchMessage.OefErrorOperation.REGISTER_SERVICE,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received oef_search error message={incoming_message} in dialogue={oef_search_dialogue}.",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the oef_search handler."""
        # setup
        incoming_message = self.build_incoming_message(
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description=self.mocked_description,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle oef_search message of performative={incoming_message.performative} in dialogue={self.oef_search_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the oef_search handler."""
        assert self.oef_search_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a the registration db model."""

import datetime
import json
import logging
import os
import sqlite3
from collections import defaultdict
from typing import Any, Dict, List, Tuple, cast

from aea.skills.base import Model


_default_logger = logging.getLogger(
    "aea.packages.fetchai.skills.confirmation_aw3.registration_db"
)


class RegistrationDB(Model):
    """Communicate between the database and the python objects."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialise the class."""
        custom_path = kwargs.pop("custom_path", None)
        super().__init__(**kwargs)
        this_dir = os.getcwd()
        self.db_path = (
            os.path.join(this_dir, "registration.db")
            if custom_path is None
            else custom_path
        )
        if not os.path.exists(os.path.dirname(os.path.abspath(self.db_path))):
            raise ValueError(f"Path={self.db_path} not valid!")  # pragma: nocover
        self._initialise_backend()

    def _initialise_backend(self) -> None:
        """Set up database and initialise the tables."""
        self._execute_single_sql(
            "CREATE TABLE IF NOT EXISTS registered_table (address TEXT NOT NULL, ethereum_address TEXT, "
            "ethereum_signature TEXT, fetchai_signature TEXT, "
            "developer_handle TEXT NOT NULL, tweet TEXT, PRIMARY KEY (address, developer_handle))"
        )
        self._execute_single_sql(
            "CREATE TABLE IF NOT EXISTS trades_table (address TEXT, created_at timestamp, data TEXT)"
        )

    def set_trade(
        self, address: str, timestamp: datetime.datetime, data: Dict[str, str],
    ) -> None:
        """Record a registration."""
        command = "INSERT INTO trades_table(address, created_at, data) values(?, ?, ?)"
        variables: Tuple[str, datetime.datetime, str] = (
            address,
            timestamp,
            json.dumps(data),
        )
        self._execute_single_sql(command, variables)

    def get_trade_count(self, address: str) -> int:
        """Get trade count."""
        command = "SELECT COUNT(*) FROM trades_table where address=?"
        ret = self._execute_single_sql(command, (address,))
        return int(ret[0][0])

    def get_developer_handle(self, address: str) -> str:
        """Get developer handle for address."""
        command = "SELECT developer_handle FROM registered_table where address=?"
        ret = self._execute_single_sql(command, (address,))
        if len(ret[0]) != 1:
            raise ValueError(
                f"More than one developer_handle found for address={address}."
            )
        return ret[0][0]

    def get_addresses(self, developer_handle: str) -> List[str]:
        """Get addresses for developer handle."""
        command = "SELECT address FROM registered_table where developer_handle=?"
        ret = self._execute_single_sql(command, (developer_handle,))
        addresses = [address[0] for address in ret]
        if len(addresses) == 0:
            raise ValueError(
                f"Should find at least one address for developer_handle={developer_handle}."
            )
        return addresses

    def get_handle_and_trades(self, address: str) -> Tuple[str, int]:
        """Get developer and number of trades for address."""
        developer_handle = self.get_developer_handle(address)
        addresses = self.get_addresses(developer_handle)
        trades = 0
        for address_ in addresses:
            trades += self.get_trade_count(address_)
        return (developer_handle, trades)

    def get_all_addresses_and_handles(self) -> List[Tuple[str, str]]:
        """Get all addresses."""
        command = "SELECT address, developer_handle FROM registered_table"
        results = cast(List[Tuple[str, str]], self._execute_single_sql(command, ()))
        return results

    def get_leaderboard(self) -> List[Tuple[str, str, int]]:
        """Get the leader board."""
        addresses_and_handles = self.get_all_addresses_and_handles()
        results_dir: Dict[Tuple[str, str], int] = defaultdict(int)
        for address, developer_handle in addresses_and_handles:
            trades = self.get_trade_count(address)
            if trades == 0:
                continue
            results_dir[(address, developer_handle)] += trades
        results = [(k[0], k[1], v) for k, v in results_dir.items()]
        results.sort(key=lambda x: x[2], reverse=True)
        return results

    def set_registered(self, address: str, developer_handle: str) -> None:
        """Record a registration."""
        if self.is_registered(address):
            return
        command = "INSERT OR REPLACE INTO registered_table(address, ethereum_address, ethereum_signature, fetchai_signature, developer_handle, tweet) values(?, ?, ?, ?, ?, ?)"
        variables = (
            address,
            "",
            "",
            "",
            developer_handle,
            "",
        )
        self._execute_single_sql(command, variables)

    def is_registered(self, address: str) -> bool:
        """Check if an address is registered."""
        command = "SELECT * FROM registered_table WHERE address=?"
        variables = (address,)
        result = self._execute_single_sql(command, variables)
        return len(result) != 0

    def _execute_single_sql(
        self,
        command: str,
        variables: Tuple[Any, ...] = (),
        print_exceptions: bool = True,
    ) -> List[Tuple[str, ...]]:
        """Query the database - all the other functions use this under the hood."""
        conn = None
        ret: List[Tuple[str, ...]] = []
        try:
            conn = sqlite3.connect(self.db_path, timeout=300)  # 5 mins
            c = conn.cursor()
            c.execute(command, variables)
            ret = c.fetchall()
            conn.commit()
        except Exception as e:  # pragma: nocover # pylint: disable=broad-except
            if print_exceptions:
                self.context.logger.warning(f"Exception in database: {e}")
        finally:
            if conn is not None:
                conn.close()

        return ret



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for aea/components/loader.py"""
from pathlib import Path
from typing import cast
from unittest import mock
from unittest.mock import Mock

import pytest
from click.exceptions import ClickException

from aea.components.loader import load_component_from_config
from aea.configurations.base import ProtocolConfig, SkillConfig
from aea.configurations.data_types import ComponentType
from aea.configurations.loader import load_component_configuration
from aea.exceptions import (
    AEAComponentLoadException,
    AEAInstantiationException,
    AEAPackageLoadingError,
)
from aea.helpers.base import cd
from aea.protocols.base import Protocol
from aea.test_tools.test_cases import AEATestCaseEmpty


@pytest.fixture(scope="module")
def component_configuration():
    """Return a component configuration."""
    return ProtocolConfig("a_protocol", "an_author", "0.1.0")


def test_component_loading_generic_exception(component_configuration):
    """Test 'load_component_from_config' method when a generic "Exception" occurs."""

    with mock.patch.object(
        Protocol, "from_config", side_effect=Exception("Generic exception")
    ):
        with pytest.raises(
            Exception, match="Package loading error: An error occurred while loading"
        ):
            load_component_from_config(component_configuration)


def test_component_loading_generic_module_not_found_error(component_configuration):
    """Test 'load_component_from_config' method when a generic "ModuleNotFoundError" occurs."""

    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError(
            "Package loading error: An error occurred while loading .*: Generic error"
        ),
    ):
        with pytest.raises(ModuleNotFoundError, match="Generic error"):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_non_framework_package(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a generic import path (non framework related."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError("No module named 'generic.package'"),
    ):
        with pytest.raises(ModuleNotFoundError):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_framework_package(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') but for some reason it doesn't contain the author name."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError("No module named 'packages'"),
    ):
        with pytest.raises(ModuleNotFoundError, match="No module named 'packages'"):
            load_component_from_config(component_configuration)


def test_component_loading_module_not_found_error_framework_package_with_wrong_author(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') with wrong author."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError("No module named 'packages.some_author'"),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match="An error occurred while loading protocol an_author/a_protocol:0.1.0:",
        ) as e:
            load_component_from_config(component_configuration)
            assert "some_type" in str(e)


def test_component_loading_module_not_found_error_framework_package_with_wrong_type(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') with correct author but wrong type."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError(
            "No module named 'packages.some_author.some_type'"
        ),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match="An error occurred while loading protocol an_author/a_protocol:0.1.0:",
        ) as e:
            load_component_from_config(component_configuration)
            assert (
                "package 'packages/some_author' of type 'protocols' exists, but cannot find module 'some_type'"
                in e.value.args[0]
            )


def test_component_loading_module_not_found_error_framework_package_with_wrong_name(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') with correct author and type but wrong name."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError(
            "No module named 'packages.some_author.protocols.some_name'"
        ),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match="An error occurred while loading protocol an_author/a_protocol:0.1.0:",
        ) as e:
            load_component_from_config(component_configuration)
        assert (
            " No AEA package found with author name 'some_author', type 'protocols', name 'some_name'"
            in e.value.args[0]
        )


def test_component_loading_module_not_found_error_framework_package_with_wrong_suffix(
    component_configuration,
):
    """Test 'load_component_from_config' method when a "ModuleNotFoundError" occurs for a framework-related import (starts with 'packages') with correct author and type but wrong suffix."""
    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=ModuleNotFoundError(
            "No module named 'packages.some_author.protocols.some_name.some_subpackage'"
        ),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match="An error occurred while loading protocol an_author/a_protocol:0.1.0:",
        ) as e:
            load_component_from_config(component_configuration)
        assert (
            "package 'packages/some_author' of type 'protocols' exists, but cannot find module 'some_subpackage'"
            in e.value.args[0]
        )


def test_component_loading_instantiation_exception(component_configuration):
    """Test 'load_component_from_config' method when a generic "Exception" occurs."""

    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=AEAInstantiationException("Generic exception"),
    ):
        with pytest.raises(AEAInstantiationException):
            load_component_from_config(component_configuration)


def test_component_loading_component_exception(component_configuration):
    """Test 'load_component_from_config' method when a generic "Exception" occurs."""

    with mock.patch.object(
        Protocol,
        "from_config",
        side_effect=AEAComponentLoadException("Generic exception"),
    ):
        with pytest.raises(
            AEAPackageLoadingError,
            match="Package loading error: An error occurred while loading protocol an_author/a_protocol:0.1.0: Generic exception",
        ):
            load_component_from_config(component_configuration)


class TestLoadFailedCauseImportedPackageNotFound(AEATestCaseEmpty):
    """Test package not found in import."""

    def test_load_component_failed_cause_package_not_found(self):
        """Test package not found in import."""
        self.add_item("skill", "fetchai/echo:latest", local=True)

        with cd(self._get_cwd()):
            echo_dir = "./vendor/fetchai/skills/echo"
            handlers_file = Path(echo_dir) / "handlers.py"
            assert handlers_file.exists()
            file_data = handlers_file.read_text()
            file_data = file_data.replace(
                "from packages.fetchai.protocols.default",
                "from packages.fetchai.protocols.not_exist_protocol",
            )
            handlers_file.write_text(file_data)
            with cd("./vendor/fetchai"):
                self.run_cli_command("fingerprint", "skill", "fetchai/echo:0.15.0")
            agent_context = Mock()
            agent_context.agent_name = self.agent_name
            configuration = cast(
                SkillConfig,
                load_component_configuration(ComponentType.SKILL, Path(echo_dir)),
            )
            configuration.directory = Path(echo_dir)
            with pytest.raises(
                ClickException,
                match=r"Package loading error: An error occurred while loading skill fetchai/echo:.*\nTraceback",
            ) as e:
                self.run_cli_command("run", cwd=self._get_cwd())
            assert "No AEA package found with author name" in str(e)
            assert "not_exist_protocol" in str(e)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the decision maker class."""

from typing import Any, Dict

from aea.common import Address
from aea.crypto.wallet import Wallet
from aea.decision_maker.base import DecisionMakerHandler as BaseDecisionMakerHandler
from aea.helpers.transaction.base import SignedMessage, SignedTransaction
from aea.identity.base import Identity
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue


class DecisionMakerHandler(BaseDecisionMakerHandler):
    """This class implements the decision maker."""

    # pylint: disable=import-outside-toplevel
    from packages.fetchai.protocols.signing.dialogues import (  # noqa: F811
        SigningDialogue,
    )
    from packages.fetchai.protocols.signing.dialogues import (  # noqa: F811
        SigningDialogues as BaseSigningDialogues,
    )
    from packages.fetchai.protocols.signing.message import SigningMessage  # noqa: F811

    class SigningDialogues(BaseSigningDialogues):
        """This class keeps track of all oef_search dialogues."""

        def __init__(self, self_address: Address, **kwargs: Any) -> None:
            """
            Initialize dialogues.

            :param self_address: the address of the entity for whom dialogues are maintained
            :return: None
            """

            def role_from_first_message(  # pylint: disable=unused-argument
                message: Message, receiver_address: Address
            ) -> BaseDialogue.Role:
                """Infer the role of the agent from an incoming/outgoing first message

                :param message: an incoming/outgoing first message
                :param receiver_address: the address of the receiving agent
                :return: The role of the agent
                """
                from packages.fetchai.protocols.signing.dialogues import (  # pylint: disable=import-outside-toplevel
                    SigningDialogue,
                )

                return SigningDialogue.Role.DECISION_MAKER

            # pylint: disable=import-outside-toplevel
            from packages.fetchai.protocols.signing.dialogues import (
                SigningDialogues as BaseSigningDialogues,
            )

            BaseSigningDialogues.__init__(
                self,
                self_address=self_address,
                role_from_first_message=role_from_first_message,
                **kwargs,
            )

    signing_dialogue_class = SigningDialogue
    signing_msg_class = SigningMessage

    def __init__(self, identity: Identity, wallet: Wallet) -> None:
        """
        Initialize the decision maker.

        :param identity: the identity
        :param wallet: the wallet
        """
        kwargs: Dict[str, Any] = {}
        super().__init__(
            identity=identity, wallet=wallet, **kwargs,
        )
        self.signing_dialogues = DecisionMakerHandler.SigningDialogues(
            self.self_address
        )

    def handle(self, message: Message) -> None:
        """
        Handle an internal message from the skills.

        :param message: the internal message
        :return: None
        """
        if isinstance(message, self.signing_msg_class):
            self._handle_signing_message(message)
        else:  # pragma: no cover
            self.logger.error(
                "[{}]: cannot handle message={} of type={}".format(
                    self.agent_name, message, type(message)
                )
            )

    def _handle_signing_message(self, signing_msg: SigningMessage) -> None:
        """
        Handle a signing message.

        :param signing_msg: the transaction message
        :return: None
        """
        signing_dialogue = self.signing_dialogues.update(signing_msg)  # type: ignore
        if signing_dialogue is None or not isinstance(
            signing_dialogue, self.signing_dialogue_class
        ):  # pragma: no cover
            self.logger.error(
                "[{}]: Could not construct signing dialogue. Aborting!".format(
                    self.agent_name
                )
            )
            return

        if signing_msg.performative == self.signing_msg_class.Performative.SIGN_MESSAGE:
            self._handle_message_signing(signing_msg, signing_dialogue)
        elif (
            signing_msg.performative
            == self.signing_msg_class.Performative.SIGN_TRANSACTION
        ):
            self._handle_transaction_signing(signing_msg, signing_dialogue)
        else:  # pragma: no cover
            self.logger.error(
                "[{}]: Unexpected transaction message performative".format(
                    self.agent_name
                )
            )

    def _handle_message_signing(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle a message for signing.

        :param signing_msg: the signing message
        :param signing_dialogue: the signing dialogue
        :return: None
        """
        performative = self.signing_msg_class.Performative.ERROR
        kwargs = {
            "error_code": self.signing_msg_class.ErrorCode.UNSUCCESSFUL_MESSAGE_SIGNING,
        }  # type: Dict[str, Any]
        signed_message = self.wallet.sign_message(
            signing_msg.raw_message.ledger_id,
            signing_msg.raw_message.body,
            signing_msg.raw_message.is_deprecated_mode,
        )
        if signed_message is not None:
            performative = self.signing_msg_class.Performative.SIGNED_MESSAGE
            kwargs.pop("error_code")
            kwargs["signed_message"] = SignedMessage(
                signing_msg.raw_message.ledger_id,
                signed_message,
                signing_msg.raw_message.is_deprecated_mode,
            )
        signing_msg_response = signing_dialogue.reply(
            performative=performative, target_message=signing_msg, **kwargs,
        )
        self.message_out_queue.put(signing_msg_response)

    def _handle_transaction_signing(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle a transaction for signing.

        :param signing_msg: the signing message
        :param signing_dialogue: the signing dialogue
        :return: None
        """
        performative = self.signing_msg_class.Performative.ERROR
        kwargs = {
            "error_code": self.signing_msg_class.ErrorCode.UNSUCCESSFUL_TRANSACTION_SIGNING,
        }  # type: Dict[str, Any]
        signed_tx = self.wallet.sign_transaction(
            signing_msg.raw_transaction.ledger_id, signing_msg.raw_transaction.body
        )
        if signed_tx is not None:
            performative = self.signing_msg_class.Performative.SIGNED_TRANSACTION
            kwargs.pop("error_code")
            kwargs["signed_transaction"] = SignedTransaction(
                signing_msg.raw_transaction.ledger_id, signed_tx
            )
        signing_msg_response = signing_dialogue.reply(
            performative=performative, target_message=signing_msg, **kwargs,
        )
        self.message_out_queue.put(signing_msg_response)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the crypto/helpers module."""
import logging
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import mock_open, patch

import pytest
from aea_ledger_cosmos import CosmosCrypto
from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_fetchai import FetchAICrypto

from aea.crypto.helpers import (
    create_private_key,
    get_wallet_from_agent_config,
    make_certificate,
    private_key_verify_or_create,
    try_generate_testnet_wealth,
    try_validate_private_key_path,
)
from aea.crypto.wallet import Wallet

from tests.conftest import (
    COSMOS_PRIVATE_KEY_FILE,
    CUR_PATH,
    ETHEREUM_PRIVATE_KEY_FILE,
    ETHEREUM_PRIVATE_KEY_PATH,
    FETCHAI_PRIVATE_KEY_PATH,
)
from tests.test_cli.tools_for_testing import AgentConfigMock


logger = logging.getLogger(__name__)


class ResponseMock:
    """Mock for response class."""

    text = "some text"

    def __init__(self, status_code=200):
        """Initialise response mock."""
        self.status_code = status_code


class TestHelperFile:
    """Test helper module in aea/crypto."""

    def tests_private_keys(self):
        """Test the private keys."""
        try_validate_private_key_path(
            FetchAICrypto.identifier, FETCHAI_PRIVATE_KEY_PATH
        )
        with pytest.raises(SystemExit):
            private_key_path = os.path.join(
                CUR_PATH, "data", "fet_private_key_wrong.txt"
            )
            try_validate_private_key_path(FetchAICrypto.identifier, private_key_path)

        try_validate_private_key_path(
            EthereumCrypto.identifier, ETHEREUM_PRIVATE_KEY_PATH
        )
        with pytest.raises(SystemExit):
            private_key_path = os.path.join(
                CUR_PATH, "data", "fet_private_key_wrong.txt"
            )
            try_validate_private_key_path(EthereumCrypto.identifier, private_key_path)

    def tests_generate_wealth_ethereum_fail_no_url(self, caplog):
        """Test generate wealth for ethereum."""
        address = "my_address"
        with caplog.at_level(
            logging.DEBUG, logger="aea_ledger_ethereum._default_logger"
        ):
            try_generate_testnet_wealth(
                identifier=EthereumCrypto.identifier, address=address
            )
            assert (
                "Url is none, no default url provided. Please provide a faucet url."
                in caplog.text
            )

    def tests_generate_wealth_ethereum_fail_invalid_url(self, caplog):
        """Test generate wealth for ethereum."""
        address = "my_address"
        result = ResponseMock(status_code=500)
        with patch("aea_ledger_ethereum.requests.get", return_value=result):
            with caplog.at_level(
                logging.DEBUG, logger="aea_ledger_ethereum._default_logger"
            ):
                try_generate_testnet_wealth(
                    identifier=EthereumCrypto.identifier,
                    address=address,
                    url="wrong_url",
                )
                assert "Response: 500" in caplog.text

    def tests_generate_wealth_ethereum_fail_valid_url(self, caplog):
        """Test generate wealth for ethereum."""
        address = "my_address"
        result = ResponseMock(status_code=200)
        with patch("aea_ledger_ethereum.requests.get", return_value=result):
            with caplog.at_level(
                logging.DEBUG, logger="aea_ledger_ethereum._default_logger"
            ):
                try_generate_testnet_wealth(
                    identifier=EthereumCrypto.identifier,
                    address=address,
                    url="correct_url",
                )

    @patch("aea_ledger_ethereum.requests.post", return_value=ResponseMock())
    @patch("aea_ledger_ethereum.json.loads", return_value={"error_message": ""})
    def test_try_generate_testnet_wealth_error_resp_ethereum(self, *mocks):
        """Test try_generate_testnet_wealth error_resp."""
        try_generate_testnet_wealth(EthereumCrypto.identifier, "address")

    def test_try_validate_private_key_path_positive(self):
        """Test _validate_private_key_path positive result."""
        try_validate_private_key_path(
            FetchAICrypto.identifier, FETCHAI_PRIVATE_KEY_PATH
        )
        try_validate_private_key_path(
            EthereumCrypto.identifier, ETHEREUM_PRIVATE_KEY_PATH
        )

    @patch("builtins.open", mock_open())
    def test__create_ethereum_private_key_positive(self, *mocks):
        """Test _create_ethereum_private_key positive result."""
        create_private_key(EthereumCrypto.identifier, ETHEREUM_PRIVATE_KEY_FILE)

    @patch("builtins.open", mock_open())
    def test__create_cosmos_private_key_positive(self, *mocks):
        """Test _create_cosmos_private_key positive result."""
        create_private_key(CosmosCrypto.identifier, COSMOS_PRIVATE_KEY_FILE)


def test_private_key_verify_or_create():
    """Test private_key_verify_or_create."""
    agent_conf = AgentConfigMock()
    with patch("aea.crypto.helpers.create_private_key") as mock_create:
        private_key_verify_or_create(agent_conf, Path("."))
    mock_create.assert_called()

    agent_conf = AgentConfigMock(private_key_paths=[("fetchai", "test")])
    with patch("aea.crypto.helpers.try_validate_private_key_path") as mock_validate:
        private_key_verify_or_create(agent_conf, Path("."))
    mock_validate.assert_called()

    agent_conf = AgentConfigMock(private_key_paths=[("fetchai", "${var}")])
    with patch("aea.crypto.helpers.try_validate_private_key_path") as mock_validate:
        with patch("aea.crypto.helpers.create_private_key") as mock_create:
            private_key_verify_or_create(agent_conf, Path("."))
    mock_validate.assert_not_called()
    mock_create.assert_not_called()


def test_make_certificate():
    """Test make_certificate."""
    with TemporaryDirectory() as tmp_dir:
        make_certificate(
            "fetchai",
            os.path.join(CUR_PATH, "data", "fetchai_private_key.txt"),
            b"message",
            os.path.join(tmp_dir, "test.txt"),
        )


def test_get_wallet_from_agent_config():
    """Test get_wallet_from_agent_config."""
    agent_conf = AgentConfigMock()
    wallet = get_wallet_from_agent_config(agent_conf)
    assert isinstance(wallet, Wallet)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the simple oracle client skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.contracts.fet_erc20.contract import PUBLIC_ID as ERC20_PUBLIC_ID
from packages.fetchai.contracts.oracle_client.contract import (
    PUBLIC_ID as CLIENT_CONTRACT_PUBLIC_ID,
)
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.skills.simple_oracle_client.behaviours import (
    SimpleOracleClientBehaviour,
)
from packages.fetchai.skills.simple_oracle_client.strategy import Strategy

from tests.conftest import ROOT_DIR


DEFAULT_ADDRESS = "0x0000000000000000000000000000000000000000"


class TestSkillBehaviour(BaseSkillTestCase):
    """Test behaviours of simple oracle client."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "simple_oracle_client"
    )

    @classmethod
    def setup(cls, **kwargs):
        """Setup the test class."""
        super().setup()
        cls.simple_oracle_client_behaviour = cast(
            SimpleOracleClientBehaviour,
            cls._skill.skill_context.behaviours.simple_oracle_client_behaviour,
        )

    def test_setup(self):
        """Test the setup method of the simple_oracle_client behaviour."""
        strategy = cast(Strategy, self.simple_oracle_client_behaviour.context.strategy)
        strategy.oracle_contract_address = DEFAULT_ADDRESS
        strategy.erc20_address = DEFAULT_ADDRESS
        strategy.is_oracle_contract_set = True

        self.simple_oracle_client_behaviour.setup()

        self.assert_quantity_in_outbox(1)

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            contract_id=str(CLIENT_CONTRACT_PUBLIC_ID),
            callable="get_deploy_transaction",
        )
        assert has_attributes, error_str

    def test_setup_with_contract_set(self):
        """Test the setup method of the simple_oracle_client behaviour for existing contract."""
        strategy = cast(Strategy, self.simple_oracle_client_behaviour.context.strategy)
        strategy.client_contract_address = DEFAULT_ADDRESS
        strategy.oracle_contract_address = DEFAULT_ADDRESS
        strategy.erc20_address = DEFAULT_ADDRESS
        strategy.is_client_contract_deployed = True
        strategy.is_oracle_contract_set = True

        with patch.object(
            self.simple_oracle_client_behaviour.context.logger, "log"
        ) as mock_logger:
            self.simple_oracle_client_behaviour.setup()
        mock_logger.assert_any_call(
            logging.INFO, "Fetch oracle client contract address already added",
        )
        self.assert_quantity_in_outbox(0)

    def test_act_pre_deploy(self):
        """Test the act method of the simple_oracle_client behaviour before contract is deployed."""

        with patch.object(
            self.simple_oracle_client_behaviour.context.logger, "log"
        ) as mock_logger:
            self.simple_oracle_client_behaviour.act()
        mock_logger.assert_any_call(
            logging.INFO, "Oracle client contract not yet deployed",
        )
        self.assert_quantity_in_outbox(0)

    def test_act_approve_transactions(self):
        """Test the act method of the simple_oracle_client behaviour before transactions are approved."""
        strategy = cast(Strategy, self.simple_oracle_client_behaviour.context.strategy)
        strategy.client_contract_address = DEFAULT_ADDRESS
        strategy.oracle_contract_address = DEFAULT_ADDRESS
        strategy.erc20_address = DEFAULT_ADDRESS
        strategy.is_client_contract_deployed = True
        strategy.is_oracle_contract_set = True

        self.simple_oracle_client_behaviour.act()
        self.assert_quantity_in_outbox(1)

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            contract_id=str(ERC20_PUBLIC_ID),
            contract_address=strategy.client_contract_address,
            callable="get_approve_transaction",
        )
        assert has_attributes, error_str

    def test_act_query(self):
        """Test the act method of the simple_oracle_client behaviour for normal querying."""
        strategy = cast(Strategy, self.simple_oracle_client_behaviour.context.strategy)
        strategy.client_contract_address = DEFAULT_ADDRESS
        strategy.oracle_contract_address = DEFAULT_ADDRESS
        strategy.erc20_address = DEFAULT_ADDRESS
        strategy.is_client_contract_deployed = True
        strategy.is_oracle_transaction_approved = True
        strategy.is_oracle_contract_set = True

        self.simple_oracle_client_behaviour.act()
        self.assert_quantity_in_outbox(1)
        assert strategy.is_oracle_contract_set

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            contract_id=str(CLIENT_CONTRACT_PUBLIC_ID),
            contract_address=strategy.client_contract_address,
            callable="get_query_transaction",
        )
        assert has_attributes, error_str

    def test_teardown(self):
        """Test that the teardown method of the simple_oracle_client behaviour leaves no messages in the outbox."""
        assert self.simple_oracle_client_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)
